// 002 u_c_cds_table
// 16 oct 2004

// -- (C) John COLIBRI 2006
// -- documentation: http://www.jcolibri.com

// -- c_data
// --   c_session
// --     c_session_list

(*$r+*)

unit u_c_cds_table;
  interface
    uses Classes, u_c_basic_object;

    type c_column= // one "column"
                   Class(c_basic_object)
                     // -- m_name:

                     Constructor create_column(p_name: String);
                     function f_display_column: String;
                     Destructor Destroy; Override;
                   end; // c_column

         c_table= Class; // forward
         c_line= // one "line"
                 Class(c_basic_object)
                   // -- m_name: the parameters of "<ROW ... >"

                   m_c_parent_table: c_table;

                   m_c_column_list: tStringList;
                   m_line_number, m_previous_line_number, m_update_code: Integer;
                   m_row_state: Integer;

                   Constructor create_line(p_name: String;
                       p_c_parent_table: c_table;
                       p_line_number, p_previous_line_number, p_update_code: Integer);

                   function f_display_line: String;

                   function f_column_count: Integer;
                   function f_c_column(p_column_index: Integer): c_column;
                   function f_index_of(p_column_name: String): Integer;
                   function f_c_find_by_column(p_column_name: String): c_column;
                   procedure add_column(p_column_name: String; p_c_element: c_column);
                   function f_c_add_column(p_column_name: String): c_column;
                   procedure display_column_list;

                   Destructor Destroy; Override;
                 end; // c_line

         c_table= // "line" list
                  Class(c_basic_object)
                    m_c_line_list: tStringList;

                    // -- puts in c_line 0 the headers
                    m_c_column_name_list: tStringList;

                    m_trace_table: Boolean;

                    Constructor create_table(p_name: String);

                    function f_line_count: Integer;
                    function f_c_line(p_line_index: Integer): c_line;
                    function f_index_of(p_line_name: String): Integer;
                    function f_c_find_by_line(p_line_name: String): c_line;
                    procedure add_line(p_line_name: String; p_c_line: c_line);
                    function f_c_add_line(p_row_number: Integer; p_log_list_string, p_parameters: String): c_line;

                    procedure display_line_list;
                    procedure display_formatted_line_list;

                    Destructor Destroy; Override;
                  end; // c_table

  implementation
    uses SysUtils, Math
        , u_strings
        , u_c_display;

    var s_trace_table: Boolean= False;

    procedure trace_table(p_text: String);
      begin
        if s_trace_table
          then display(p_text);
      end; // trace_table

    // -- c_column

    Constructor c_column.create_column(p_name: String);
      begin
        Inherited create_basic_object(p_name);
      end; // create_column

    function c_column.f_display_column: String;
      begin
        Result:= Format('%-10s ', [m_name]);
      end; // f_display_column

    Destructor c_column.Destroy;
      begin
        InHerited;
      end; // Destroy

    // -- c_line

    Constructor c_line.create_line(p_name: String;
        p_c_parent_table: c_table;
        p_line_number, p_previous_line_number, p_update_code: Integer);

      procedure analyze_columns;
          // -- F_NUMERO="5" F_NOM="UML et Patterns Delphi " F_JOURS="3" F_PRIX="1400.00"
        var l_column_index: Integer;

        procedure add_column(p_key, p_value: String);
          var l_key_column_index: Integer;
          begin
            // -- add the column name if not yet known
            with m_c_parent_table do
            begin
              l_key_column_index:= m_c_column_name_list.IndexOf(p_key);
              if l_key_column_index< 0
                then begin
                    // -- add a new column name
                    l_key_column_index:= m_c_column_name_list.Count;
                    m_c_column_name_list.Add(p_key);
                  end;
            end; // with m_c_parent_table

            // -- now add "nnn=vvv"
            f_c_add_column(IntToStr(l_key_column_index)+ '='+ p_value);
          end; // add_column

        var l_string_index: Integer;
            l_key, l_value: String;

        begin // create_line
          trace_table(' >'+ m_name+ '<');
          l_string_index:= 1;
          l_column_index:= 0;

          while l_string_index<= Length(m_name) do
          begin
            l_key:= f_string_extract_identifier(m_name, l_string_index);
            f_skip_spaces(m_name, l_string_index);
            // -- skip "="
            Inc(l_string_index);
            l_value:= f_string_extract_quoted_string(m_name, l_string_index);
            trace_table('  '+ l_key+ ' | '+ l_value);

            if l_key= 'RowState'
              then m_row_state:= StrToInt(l_value)
              else begin
                  add_column(l_key, l_value);
                  Inc(l_column_index);
                end;

            if (l_string_index<= Length(m_name)) and (m_name[l_string_index]= '/')
              then Break;
          end; // while l_string_index
        end; // analyze_columns

      begin // create_line
        Inherited create_basic_object(p_name);

        m_c_parent_table:= p_c_parent_table;

        m_line_number:= p_line_number;
        m_previous_line_number:= p_previous_line_number;
        m_update_code:= p_update_code;

        m_c_column_list:= tStringList.Create;

        // -- analyze the content of the "<ROW ... " parameters
        analyze_columns;
      end; // create_line

    function c_line.f_display_line: String;
      begin
        Result:= Format('%3d [%3d %3d] ', [m_line_number, m_previous_line_number, m_update_code])+ m_name;
      end; // f_display_line

    function c_line.f_column_count: Integer;
      begin
        Result:= m_c_column_list.Count;
      end; // f_column_count

    function c_line.f_c_column(p_column_index: Integer): c_column;
      begin
        Result:= c_column(m_c_column_list.Objects[p_column_index]);
      end; //  f_c_column

    function c_line.f_index_of(p_column_name: String): Integer;
      begin
        Result:= m_c_column_list.IndexOf(p_column_name);
      end; // f_index_of

    function c_line.f_c_find_by_column(p_column_name: String): c_column;
      var l_index_of: Integer;
      begin
        l_index_of:= f_index_of(p_column_name);
        if l_index_of< 0
          then Result:= Nil
          else Result:= c_column(m_c_column_list.Objects[l_index_of]);
      end; // f_c_find_by_column

    procedure c_line.add_column(p_column_name: String; p_c_element: c_column);
      begin
        m_c_column_list.AddObject(p_column_name, p_c_element);
      end; // add_column

    function c_line.f_c_add_column(p_column_name: String): c_column;
      begin
        Result:= f_c_column(m_c_column_list.AddObject(p_column_name, c_column.create_column(p_column_name)));
      end; // f_c_add_column

    procedure c_line.display_column_list;
      var l_column_index: Integer;
      begin
        display(m_name+ ' '+ IntToStr(f_column_count));

        for l_column_index:= 0 to f_column_count- 1 do
          display(f_c_column(l_column_index).f_display_column);
      end; // display_column_list

    Destructor c_line.Destroy;
      var l_column_index: Integer;
      begin
        for l_column_index:= 0 to f_column_count- 1 do
          f_c_column(l_column_index).Free;
        m_c_column_list.Free;

        InHerited;
      end; // Destroy

    // -- c_table

    Constructor c_table.create_table(p_name: String);
      begin
        Inherited create_basic_object(p_name);

        s_trace_table:= m_trace_table;
        m_c_column_name_list:= tStringList.Create;
        m_c_line_list:= tStringList.Create;
      end; // create_table

    function c_table.f_line_count: Integer;
      begin
        Result:= m_c_line_list.Count;
      end; // f_line_count

    function c_table.f_c_line(p_line_index: Integer): c_line;
      begin
        Result:= c_line(m_c_line_list.Objects[p_line_index]);
      end; //  f_c_line

    function c_table.f_index_of(p_line_name: String): Integer;
      begin
        Result:= m_c_line_list.IndexOf(p_line_name);
      end; // f_index_of

    function c_table.f_c_find_by_line(p_line_name: String): c_line;
      var l_index_of: Integer;
      begin
        l_index_of:= f_index_of(p_line_name);
        if l_index_of< 0
          then Result:= Nil
          else Result:= c_line(m_c_line_list.Objects[l_index_of]);
      end; // f_c_find_by_name

    procedure c_table.add_line(p_line_name: String; p_c_line: c_line);
      begin
        m_c_line_list.AddObject(p_line_name, p_c_line);
      end; // add_line

    function c_table.f_c_add_line(p_row_number: Integer; p_log_list_string, p_parameters: String): c_line;
      var l_string_index: Integer;
          l_line_number, l_previous_line_number, l_update_code: Integer;
      begin
        if Trim(p_log_list_string)= ''
          then Result:= c_line.create_line(p_parameters, Self, p_row_number, -1, -1)
          else begin
              // -- analyze the p_log_list_string
              l_string_index:= 1;
              l_line_number:= f_string_extract_integer(p_log_list_string, l_string_index);
              l_previous_line_number:= f_string_extract_integer(p_log_list_string, l_string_index);
              l_update_code:= f_string_extract_integer(p_log_list_string, l_string_index);

              Result:= c_line.create_line(p_parameters, Self, p_row_number, l_previous_line_number, l_update_code);
            end;

        add_line('', Result);
      end; // f_c_add_table_line

    procedure c_table.display_line_list;
      var l_line_index: Integer;
      begin
        display(m_name+ ' '+ IntToStr(f_line_count));

        for l_line_index:= 0 to f_line_count- 1 do
          display(f_c_line(l_line_index).f_display_line);
      end; // display_line_list

    procedure c_table.display_formatted_line_list;
      var l_oa_column_size: array of Integer;

      procedure compute_column_sizes;
        var l_column_index: Integer;
            l_line_index: Integer;
        begin
          SetLength(l_oa_column_size, m_c_column_name_list.Count);
          for l_column_index:= 0 to m_c_column_name_list.Count- 1 do
            l_oa_column_size[l_column_index]:= Length(m_c_column_name_list[l_column_index]);

          for l_line_index:= 0 to f_line_count- 1 do
            with f_c_line(l_line_index) do
                for l_column_index:= 0 to f_column_count- 1 do
                  l_oa_column_size[l_column_index]:= Max(l_oa_column_size[l_column_index],
                      Length(m_c_column_list.Values[IntToStr(l_column_index)]));
        end; // compute_column_sizes

      procedure display_values;
        //   TUpdateStatus = (usUnmodified, usModified, usInserted, usDeleted);

        function f_display_triplet_row_state(p_previous_line_number,
            p_update_code, p_row_state: Integer): String;
          begin
            if p_previous_line_number<= 0
              then Result:= '[    '
              else Result:= Format('[%3d ', [p_previous_line_number]);

            case p_update_code of
              -1 : Result:= Result+ '    ';
              4  : Result:= Result+ 'add ';
              8  : Result:= Result+ 'del ';
              else Result:= Result + Format('%3d ', [p_update_code]);
            end;

            case p_row_state of
              4  : Result:= Result+ 'R_add  ';
              6  : Result:= Result+ 'R_del  ';
              5  : Result:= Result+ 'R_base ';
              13 : Result:= Result+ 'R_midd ';
              12 : Result:= Result+ 'R_last ';
              else Result:= Result + Format('R%6d', [p_row_state]);
            end; // case p_row_state

            Result:= Result+ ']';
          end; // f_display_triplet_row_state

        var l_column_index: Integer;
            l_line: String;
            l_value: String;
            l_line_index: Integer;
        begin
          l_line:= f_spaces(21);
          for l_column_index:= 0 to m_c_column_name_list.Count- 1 do
            l_line:= l_line+ Format(' | %-'+ IntToStr(l_oa_column_size[l_column_index])+ 's',
                [m_c_column_name_list[l_column_index]]);
          display(l_line);
          display_line;

          for l_line_index:= 0 to f_line_count- 1 do
            with f_c_line(l_line_index) do
              begin
                l_line:= Format('%3d ', [m_line_number+ 1])
                  + f_display_triplet_row_state(m_previous_line_number, m_update_code, m_row_state);

                for l_column_index:= 0 to m_c_column_name_list.Count- 1 do
                begin
                  l_value:= m_c_column_list.Values[IntToStr(l_column_index)];
                  l_line:= l_line+ Format(' | %-'+ IntToStr(l_oa_column_size[l_column_index])+ 's',
                      [l_value]);
                end;

                display(l_line);
              end;
        end; // display_values

      begin // display_formatted_line_list
        compute_column_sizes;
        display_values;
      end; // display_formatted_line_list

    Destructor c_table.Destroy;
      var l_line_index: Integer;
      begin
        for l_line_index:= 0 to f_line_count- 1 do
          f_c_line(l_line_index).Free;
        m_c_line_list.Free;
        m_c_column_name_list.Free;

        Inherited;
      end; // Destroy

    begin // u_c_line_list
    end. // u_c_line_list
